//
//  lariat-online/daq/src/LariatFragment.cc
//

#include "LariatFragment.h"
#include "RawBuffer.h"
#include <fstream>
#include <sstream>
#include <string.h>

// My three constructors:

LariatFragment::LariatFragment(std::istream & in)
{
  fillFromBuffer(in);
}

LariatFragment::LariatFragment(std::string & fileName)
{
  std::ifstream in(fileName);
  fillFromBuffer(in);
  in.close();
}

LariatFragment::LariatFragment(const char * buffer, size_t size)
{
  char * ptr = (char *)buffer;
  RawBuffer rawBuffer(ptr,size);
  std::istream in(&rawBuffer);
  fillFromBuffer(in);
}


//
// Convenience function:
//

const char* LariatFragment::fragmentNameFromType(int id)
{
  switch(id){
  case FRAGMENT_TYPE_RUNINFO:
    return "Run Info";
  case FRAGMENT_TYPE_SPILLINFO:
    return "Spill Info";
  case FRAGMENT_TYPE_SPILLTRAILER:
    return "Spill Trailer";
  case FRAGMENT_TYPE_CAEN_ADC:
    return "CAEN Digitizer";
  case FRAGMENT_TYPE_V1495:
    return "CAEN V1495";
  case FRAGMENT_TYPE_TDC_CONTROLLER_HEADER:
    return "TDC Controller Header";
  case FRAGMENT_TYPE_TDC_TDC_SPILL_HEADER:
    return "TDC Spill Header";
  case FRAGMENT_TYPE_TDC_TDC_EVENT_HEADER:
    return "TDC Event Header";
  case FRAGMENT_TYPE_WUT_HIT_HEADER:
    return "WUT Hit Header";
  case FRAGMENT_TYPE_LARIAT_CONFIG:
    return "Lariat Config";
  case FRAGMENT_TYPE_LARASIC_CONFIG:
    return "LARASIC Config";
  case FRAGMENT_TYPE_READOUT_ERROR:
    return "Readout Error";
  case FRAGMENT_TYPE_TRIGGER_FRAGMENT:
    return "Trigger Fragment";

  default:
    return "Unknown";
    break;
  }
  return "Unknown";
}


/*
uint32_t LariatFragment::streamToBuffer(std::ostream& output){

  //

  return 0;
}
*/

// Upacking functions:
  
uint32_t LariatFragment::fillFromBuffer(std::istream& input)
{
  using std::cout; using std::endl;
  int fragmentSize = 0;
  int fragmentType = 0;

  int nBytes=0;
  for ( int i=0; i<FRAGMENT_TYPE_N; i++) { count[i] = 0;}

  while ( input.good() )
  {
    std::streampos starting_position=input.tellg();
    input.read((char *)&fragmentSize,sizeof(fragmentSize));
    input.read((char *)&fragmentType,sizeof(fragmentType));
    if ( input.good() )
    {
      if ( ( fragmentType < FRAGMENT_TYPE_N ) && 
	   ( fragmentType >= 0 ) )
      { count[fragmentType]++;}
      input.seekg(starting_position);

      // Case statement doesn't work here due to scoping
      if( fragmentType ==  FRAGMENT_TYPE_RUNINFO)
      {
	nBytes += fillRunInfoFromBuffer(input);
      }
      else if( fragmentType ==  FRAGMENT_TYPE_SPILLINFO)
      {
	nBytes += fillSpillInfoFromBuffer(input);
      }
      else if( fragmentType ==  FRAGMENT_TYPE_SPILLTRAILER)
      {
	nBytes += fillSpillTrailerFromBuffer(input);
	//printSpillTrailer();
      }
      else if( fragmentType ==  FRAGMENT_TYPE_CAEN_ADC)
      {
	CAENFragment frag;
	nBytes+=frag.fillFromBuffer(input);
	caenFrags.push_back(frag);      
      }
      else if(fragmentType ==  FRAGMENT_TYPE_WUT_HIT_HEADER)
      {
      	WUTFragment frag;
      	nBytes+=frag.fillFromBuffer(input);
      	wutFrags.push_back(frag);	
      }
      else if( fragmentType ==  FRAGMENT_TYPE_V1495)
      {
	V1495Fragment frag;
	nBytes+=frag.fillFromBuffer(input);
	v1495Frags.push_back(frag);      
      }
      else if( fragmentType ==  FRAGMENT_TYPE_TDC_CONTROLLER_HEADER )
      {
	TDCFragment frag;
	nBytes += frag.fillFromBuffer(input);
	tdcFrags.push_back(frag);      
      }
      else if( fragmentType ==  FRAGMENT_TYPE_LARIAT_CONFIG )
      {
	nBytes += fillLariatConfigFromBuffer(input);
      }
      else if( fragmentType ==  FRAGMENT_TYPE_LARASIC_CONFIG )
      {
	LARASICFragment frag;
	nBytes += frag.fillFromBuffer(input);
	larasicFrags.push_back(frag);
      }
      else if ( fragmentType ==  FRAGMENT_TYPE_READOUT_ERROR )
      {
	ReadoutError error;
	nBytes += error.fillFromBuffer(input);
	errorFrags.push_back(error);
      }
      else if ( fragmentType ==  FRAGMENT_TYPE_TRIGGER_FRAGMENT )
      {
	nBytes += triggerFragment.fillFromBuffer(input);
      }
      else
      {
	std::cout << "Unrecognized fragmentType " << fragmentType << 
	  " 0x" << std::hex << fragmentType << std::dec 
		  << " of size " << fragmentSize << std::endl;
	std::streamoff fSize = fragmentSize;
	
	// Fast forward to next fragment:
	input.seekg(starting_position + fSize);
	nBytes += fragmentSize;
      }
    }
  }

  for ( int i=0; i<FRAGMENT_TYPE_N; i++)
  { std::cout << "Type " << i << " had " << count[i] << " fragments" << std::endl;}

  return nBytes;
}

uint32_t LariatFragment::fillRunInfoFromBuffer(std::istream& input)
{
  uint32_t nBytes = 0;
  input.read((char *)&runInfo,sizeof(struct RunInfo));
  nBytes += sizeof(struct RunInfo);
  return(nBytes);
}

uint32_t LariatFragment::fillSpillInfoFromBuffer(std::istream& input)
{
  uint32_t nBytes = 0;
  input.read((char *)&spillInfo,sizeof(struct SpillInfo));
  nBytes += sizeof(struct SpillInfo);
  return(nBytes);
}

uint32_t LariatFragment::fillSpillTrailerFromBuffer(std::istream& input)
{
  uint32_t nBytes = 0;
  input.read((char *)&spillTrailer,sizeof(struct SpillTrailer));
  nBytes += sizeof(struct SpillTrailer);
  return(nBytes);
}

uint32_t LariatFragment::fillLariatConfigFromBuffer(std::istream& input)
{
  uint32_t nBytes = 0;
  input.read((char *)&lariatConfig.fragmentSize, 
	     sizeof(lariatConfig.fragmentSize));
  nBytes += sizeof(lariatConfig.fragmentSize);

  input.read((char *)&lariatConfig.fragmentType, 
	     sizeof(lariatConfig.fragmentType));
  nBytes += sizeof(lariatConfig.fragmentSize);

  uint32_t remainder = lariatConfig.fragmentSize - nBytes;
  lariatConfig.xml = (char *)malloc(remainder);
  input.read(lariatConfig.xml, remainder);

  return(lariatConfig.fragmentSize);
}

void LariatFragment::printLariatConfig(std::ostream & os)
{
  os << (*this).lariatConfig;
}

void LariatFragment::printSpillTrailer(std::ostream & os)
{
  os << (*this).spillTrailer;
}

void LariatFragment::print(std::ostream & os)
{
  os << *this;
}

void LariatFragment::checkOStream(std::ostream & buffer)
{
  if ( buffer.bad() || buffer.eof() || buffer.fail() || !buffer.good() )
  {
    std::cout << "Buffer error: ";
    if ( buffer.bad() )  std::cout << "BAD ";
    if ( buffer.eof() )  std::cout << "EOF ";
    if ( buffer.fail() ) std::cout << "FAIL ";
    if ( !buffer.good() ) std::cout << "NOT_GOOD ";
    throw std::ios_base::failure("Output Stream Failure");
  }
  return;
}

std::ostream& operator<<(std::ostream& os, const LariatFragment& e)
{
  std::stringstream header;
  header << "nBytes Read " << e.nBytes << std::endl;
  for ( int i=0; i<LariatFragment::FRAGMENT_TYPE_N; i++)
  {
    header << i << ": " <<LariatFragment::fragmentNameFromType(i) 
       << " " << e.count[i] << std::endl;
  }

  std::stringstream fragments;
  for ( auto &it: e.v1495Frags )   { fragments << it << std::endl; }
  for ( auto &it: e.tdcFrags )     { fragments << it << std::endl; }
  for ( auto &it: e.larasicFrags ) { fragments << it << std::endl; }
  for ( auto &it: e.wutFrags )     { fragments << it << std::endl; }
  for ( auto &it: e.errorFrags )   { fragments << it << std::endl; }
  for ( auto &it: e.caenFrags )    { fragments << it << std::endl; }

  os << header.str()
     << e.runInfo << std::endl 
     << e.lariatConfig << std::endl
     << e.spillInfo << std::endl
     << e.triggerFragment << std::endl
     << fragments.str() 
     << e.spillTrailer << std::endl;
  return(os);
}

std::ostream& operator<<(std::ostream& os, 
			 const LariatFragment::LariatConfig& e)
{
  os << "LariatConfig"       << std::endl
     << "  FragmentSize    " << e.fragmentSize  << std::endl
     << "  FragmentType    " << e.fragmentType  << std::endl
     << "  XML:" << std::endl
     << e.xml;
  return(os);
}

std::ostream& operator<<(std::ostream& os, 
			 const LariatFragment::RunInfo& e)
{
  std::stringstream text;
  text << "  BoardIds             ";
  for (uint32_t i=0; i<e.nBoards; i++) { text << " " << e.boardId[i]; }

  os << "RunInfo Contents" << std::endl
     << "  FragmentSize          " << e.fragmentSize << std::endl
     << "  FragmentType          " << e.fragmentType << std::endl
     << "  RunNumber             " << e.runNumber << std::endl
     << "  RunType               " << e.runType << std::endl
     << "  RecordLength          " << e.recordLength << std::endl
     << "  nBuffers              " << e.nBuffers << std::endl
     << "  TimeStamp             " << e.timeStamp << std::endl
     << "  nBoards               " << e.nBoards << std::endl
     << text.str() << std::endl
     << "  Link                  " << e.link << std::endl
     << "  PostPercent           " << e.postPercent << std::endl
     << "  MaxEventsPerTransfer  " << e.maxEventsPerTransfer << std::endl
     << "  EventsPerInterrupt    " << e.eventsPerInterrupt << std::endl
     << "  IrqWaitTime           " << e.irqWaitTime << std::endl
     << "  UsePedestals          " << e.usePedestals << std::endl
     << "  DacValue              " << e.dacValue << std::endl
     << "  SamplingRate          " << e.samplingRate << std::endl
     << "  SampleReduction       " << e.sampleReduction << std::endl
     << "  TdcHostName           " << e.tdcHostName << std::endl
     << "  TdcPort               " << e.tdcPort << std::endl
     << "  nTDCs                 " << e.nTDCs << std::endl
     << "  TdcEnableMask         " << e.tdcEnableMask << std::endl
     << "  TdcPulserEnable       " << e.tdcPulserEnable << std::endl
     << "  TdcNBoards            " << e.tdcNBoards << std::endl
     << "  TdcTimeout            " << e.tdcTimeOut;
  return(os);
}


std::ostream& operator<<(std::ostream& os, 
			 const LariatFragment::SpillInfo& e)
{
  os << "SpillInfo contents" << std::endl
     << "FragmentSize            " << e.fragmentSize << std::endl
     << "FragmentType            " << e.fragmentType << std::endl
     << "RunNumber               " << e.runNumber << std::endl
     << "SpillNumber             " << e.spillNumber << std::endl
     << "TimeStamp               " << e.timeStamp;
  return(os);
}

std::ostream& operator<<(std::ostream& os, 
			 const LariatFragment::SpillTrailer& e)
{
  os << "SpillTrailer contents" << std::endl
     << "  FragmentSize    " << e.fragmentSize  << std::endl
     << "  FragmentType    " << e.fragmentType  << std::endl
     << "  RunNumber       " << e.runNumber     << std::endl
     << "  SpillNumber     " << e.spillNumber   << std::endl
     << "  TimeStamp       " << e.timeStamp     << std::endl
     << "  SpillSnapshot   " << os.precision(16)<< e.spillSnapshot << std::endl
     << "  SpillOffset     " << e.spillOffset   << std::endl
     << "  SpillV1740      " << e.spillV1740    << std::endl
     << "  SpillV1751      " << e.spillV1751    << std::endl
     << "  CountV1740      " << e.countV1740    << std::endl
     << "  CountV1751      " << e.countV1751;

  return(os);
}

